import os.path

from msbff.processing import processing_pipeline
from msbff.preprocessing import preprocessing_pipeline
from msbff.visualization import plot_pipeline
from msbff.default import M1_OUTPUT_FILENAME, M2_BLOCK_SCORE, M2_MAX_INHIBITION_RATE, M2_RELATIVE_SIGNAL_INTENSITY
from pathlib import Path


def run_bioff_pipeline(input_csv,
                       output_folder,
                       pcc_threshold,
                       reliability_threshold,
                       mz_lower,
                       mz_upper,
                       sn_threshold,
                       rt_lower,
                       rt_upper,
                       fdr_threshold,
                       rt_binning,
                       mz_binning):
    """Run Bioactive Fractions Filtering Pipeline.

    Args: See the command help information for detailed arguments descriptions.

    Returns: All results will be saved to the specified folder when the program is finished.

    """

    """Check output path."""
    if not Path(output_folder).is_dir():
        raise Exception("Wrong Output Path!")

    """Module1-Preprocessing data: Read data and perform data cleaning."""
    preprocessed_df, class_bioactivity = preprocessing_pipeline(input_csv,
                                                                pcc_threshold,
                                                                reliability_threshold,
                                                                mz_lower,
                                                                mz_upper,
                                                                sn_threshold,
                                                                rt_lower,
                                                                rt_upper,
                                                                fdr_threshold)
    """Save the results of Module1 processing."""
    preprocessed_df.to_csv(os.path.join(output_folder, M1_OUTPUT_FILENAME), index=False)

    """Module2-Processing data: Calculate each variable value."""
    block_score_df, max_inhibition_rate_df, relative_signal_intensity_df = processing_pipeline(preprocessed_df,
                                                                                               rt_lower,
                                                                                               rt_upper,
                                                                                               rt_binning,
                                                                                               mz_lower,
                                                                                               mz_upper,
                                                                                               mz_binning)
    """Save the results of Module2 processing."""
    block_score_df.to_csv(os.path.join(output_folder, M2_BLOCK_SCORE))
    max_inhibition_rate_df.to_csv(os.path.join(output_folder, M2_MAX_INHIBITION_RATE))
    relative_signal_intensity_df.to_csv(os.path.join(output_folder, M2_RELATIVE_SIGNAL_INTENSITY))

    """Module3-Data visualization: Plot one line chart and three heatmap."""
    plot_pipeline(class_bioactivity, block_score_df, max_inhibition_rate_df, relative_signal_intensity_df,
                  output_folder)



# names in input table
CLASS = "Class"
BIOACTIVITY = "Bioactivity"

RT_COL = "Average Rt(min)"
MZ_COL = "Average Mz"
SN_COL = "S/N average"
PCC_COL = "PCC"
RELI_COL = "Reliability"
FDR_COL = "FDR"

TARGET_COLUMNS = [RT_COL, MZ_COL, SN_COL]

# output file names
M1_OUTPUT_FILENAME = "data_extraction.csv"
M2_BLOCK_SCORE = "block_score.csv"
M2_MAX_INHIBITION_RATE = "max_inhibition_rate.csv"
M2_RELATIVE_SIGNAL_INTENSITY = "relative_signal_intensity.csv"
M3_FIG_NAME = "Fig.png"

# plot config
FIG_A_COLOR = "purple"
FIG_B_COLOR = "Blues"
FIG_C_COLOR = "Reds"
FIG_D_COLOR = "Greens"

FIG_A_XLABEL = "Class"
FIG_A_YLABEL = "Bioactivity"

FIG_BCD_XLABEL = "Retention Time (min)"
FIG_BCD_YLABEL = "Molecular Weight (Da)"

FIG_B_CBAR_TITILE = "Block Score"
FIG_C_CBAR_TITILE = "Maximum Inhibition Rate Per Block"
FIG_D_CBAR_TITILE = "Relative Signal Intensity Per Block"

NUMBER_TITLE_FONTDICT = {"fontsize": 28, "weight": 'bold'}
LABEL_FONTDICT = {"fontsize": 16}



import pandas as pd
from msbff.default import TARGET_COLUMNS, PCC_COL, RELI_COL, FDR_COL, MZ_COL, SN_COL, RT_COL, CLASS, BIOACTIVITY


def read_rawdata(csv_file: str) -> (pd.DataFrame, pd.DataFrame, pd.Series):
    """Read the raw data and get the target content.

    Args:
        csv_file: Raw data in csv format.

    Returns: Three pieces of data content for subsequent processing.
    """
    class_bioactivity = pd.read_csv(csv_file, nrows=1).dropna(axis=1).set_index(keys=[CLASS]).loc[BIOACTIVITY]
    bioactivity_objective = class_bioactivity.copy()
    main_df = pd.read_csv(csv_file, header=3)
    target_df = main_df[TARGET_COLUMNS]
    bioactivity_df = main_df.iloc[:, -len(bioactivity_objective):]
    bioactivity_objective.index = bioactivity_df.columns
    return target_df, bioactivity_df, bioactivity_objective, class_bioactivity


def calc_pcc(df: pd.DataFrame, objective: pd.Series) -> pd.Series:
    """Calculate the Pearson correlation coefficient for the Bioactivity term.

    Args:
        df: Dataframe of bioactivity data.
        objective: Objective item for bioactivity data.

    Returns: A Series of pairwise correlations.
    """
    pcc = df.corrwith(other=objective, method="pearson", axis=1)
    pcc.name = PCC_COL
    return pcc


def calc_reli(df: pd.DataFrame) -> pd.Series:
    """Calculate the Reliability for the Bioactivity term.

    Args:
        df: Dataframe of bioactivity data.

    Returns: A Series of pairwise reliability.
    """
    reli = 1 - (df == 0).astype(int).sum(axis=1) / len(df.columns)
    reli.name = RELI_COL
    return reli


def filter_data(df: pd.DataFrame,
                pcc_threshold: float,
                reli_threshold: float,
                mz_lower: float,
                mz_upper: float,
                sn_threshold: float,
                rt_lower: float,
                rt_upper: float
                ) -> pd.DataFrame:
    """Filter data according to limited requirements.

    Args:
        df: Data to be filtered.
        pcc_threshold: Pearson correlation coefficient threshold.
        reli_threshold: Reliability threshold.
        mz_lower: Lower limit of the average Mz.
        mz_upper: Upper limit of the average Mz.
        sn_threshold: S/N average threshold.
        rt_lower: Lower limit of the Average Rt(min).
        rt_upper: Upper limit of the Average Rt(min).

    Returns: Filtered data.
    """

    pcc_condition = (df[PCC_COL] > pcc_threshold)
    reli_condition = (df[RELI_COL] >= reli_threshold)
    mz_condition = (mz_lower < df[MZ_COL]) & (df[MZ_COL] <= mz_upper)
    sn_condition = (df[SN_COL] > sn_threshold)
    rt_condition = (rt_lower < df[RT_COL]) & (df[RT_COL] <= rt_upper)

    filtered_df = df[pcc_condition & reli_condition & mz_condition & sn_condition & rt_condition]
    return filtered_df


def calc_fdr(df: pd.DataFrame) -> pd.Series:
    """Calculate FDR for the data.

    Args:
        df: A dataframe sorted by PCC.

    Returns: A Series of FDR.

    """
    fdr = pd.Series((df.index + 1) / len(df), name=FDR_COL)
    return fdr


def fdr_filter(df: pd.DataFrame, fdr_threshold: float) -> pd.DataFrame:
    """Filter data according to FDR.

    Args:
        df: Data to be filtered.
        fdr_threshold: FDR threshold.

    Returns: Filtered data.

    """
    fdr_condition = (df[FDR_COL] >= fdr_threshold)
    filtered_df = df[fdr_condition]
    return filtered_df


def preprocessing_pipeline(csv_file: str,
                           pcc_threshold: float,
                           reli_threshold: float,
                           mz_lower: float,
                           mz_upper: float,
                           sn_threshold: float,
                           rt_lower: float,
                           rt_upper: float,
                           fdr_threshold: float
                           ):
    target_df, bioactivity_df, bioactivity_objective, class_bioactivity = read_rawdata(csv_file)
    pcc = calc_pcc(bioactivity_df, bioactivity_objective)
    reli = calc_reli(bioactivity_df)
    concat_df = pd.concat([target_df, pcc, reli], axis=1)
    filtered_df = filter_data(concat_df,
                              pcc_threshold,
                              reli_threshold,
                              mz_lower,
                              mz_upper,
                              sn_threshold,
                              rt_lower,
                              rt_upper)
    sorted_df = filtered_df.sort_values(by=[PCC_COL], ignore_index=True)
    fdr = calc_fdr(sorted_df)
    fdr_concat_df = pd.concat([sorted_df, fdr], axis=1)
    fdr_filtered_df = fdr_filter(fdr_concat_df, fdr_threshold)
    return fdr_filtered_df, class_bioactivity


# if __name__ == '__main__':
#     fp = "/Users/zhouzhenyi/Documents/github/SciProc/BioFF/msbff/test/rawdata.csv"
#     # preprocessing_pipeline(fp, 0, 0.4, 100, 700, 100, 0, 7, 0.05).to_csv("dataextraction.csv", index=False)
#     _, _, _, df = read_rawdata(fp)
#     print(df)



import numpy as np
import pandas as pd
from msbff.default import RT_COL, MZ_COL, PCC_COL, SN_COL


def calc_pcc_sum(df: pd.DataFrame):
    """calculate the summation of PCC column.

    Args:
        df: Filtered dataframe.

    Returns: Summation of PCC column.

    """
    pcc_sum = df[PCC_COL].sum()
    return pcc_sum


def calc_sn_sum(df: pd.DataFrame):
    """calculate the summation of S/N average column.

    Args:
        df: Filtered dataframe.

    Returns: Summation of S/N average column.

    """
    sn_sum = df[SN_COL].sum()
    return sn_sum


def block(df: pd.DataFrame,
          rt_lower, rt_upper, rt_binning,
          mz_lower, mz_upper, mz_binning):
    n_rt_bins = int((rt_upper - rt_lower) // rt_binning)
    n_mz_bins = int((mz_upper - mz_lower) // mz_binning)

    rt_bins = np.arange(rt_lower, rt_lower + (n_rt_bins + 1) * rt_binning, rt_binning)
    mz_bins = np.arange(mz_lower, mz_lower + (n_mz_bins + 1) * mz_binning, mz_binning)

    groups = df.groupby(by=[pd.cut(df[RT_COL], bins=rt_bins), pd.cut(df[MZ_COL], bins=mz_bins)])
    return groups


def block_score(groups, pcc_sum):
    block_score_df = groups[PCC_COL].sum().to_frame().unstack() / pcc_sum * 100
    return block_score_df


def max_inhibition_rate_per_block(groups):
    max_inhibition_rate_df = groups[PCC_COL].max().to_frame().unstack()
    return max_inhibition_rate_df


def relative_signal_intensity_per_block(groups, sn_sum):
    relative_signal_intensity_df = groups[MZ_COL].sum().to_frame().unstack() / sn_sum * 100
    return relative_signal_intensity_df


def processing_pipeline(df: pd.DataFrame, rt_lower, rt_upper, rt_binning, mz_lower, mz_upper, mz_binning):
    groups = block(df, rt_lower, rt_upper, rt_binning, mz_lower, mz_upper, mz_binning)

    pcc_sum = calc_pcc_sum(df)
    block_score_df = block_score(groups, pcc_sum)

    max_inhibition_rate_df = max_inhibition_rate_per_block(groups)

    sn_sum = calc_sn_sum(df)
    relative_signal_intensity_df = relative_signal_intensity_per_block(groups, sn_sum)

    return block_score_df, max_inhibition_rate_df, relative_signal_intensity_df

#
# if __name__ == '__main__':
#     csv_path = "/Users/zhouzhenyi/Documents/github/SciProc/BioFF/msbff/test/dataextraction.csv"
#     df = pd.read_csv(csv_path)
#     block_score_df, max_inhibition_rate_df, relative_signal_intensity_df = processing_pipeline(df, 1, 100)
#     block_score_df.to_csv("block_score.csv")
#     max_inhibition_rate_df.to_csv("max_inhibition_rate.csv")
#     relative_signal_intensity_df.to_csv("relative_signal_intensity.csv")



import os.path

import matplotlib.pyplot as plt
import numpy as np
from msbff.default import (FIG_A_COLOR, FIG_B_COLOR, FIG_C_COLOR, FIG_D_COLOR,
                           FIG_A_XLABEL, FIG_A_YLABEL, FIG_BCD_XLABEL, FIG_BCD_YLABEL,
                           FIG_B_CBAR_TITILE, FIG_C_CBAR_TITILE, FIG_D_CBAR_TITILE,
                           NUMBER_TITLE_FONTDICT, LABEL_FONTDICT, M3_FIG_NAME)

# # test import (待删)
# from preprocessing import read_rawdata
# import pandas as pd
# from processing import processing_pipeline


# # test import (待删)

def set_layout():
    fig = plt.figure(figsize=(16, 12))  # 16 * 12

    ax1 = fig.add_axes([0.1, 0.55, 0.3, 0.3])
    ax2 = fig.add_axes([0.45, 0.55, 0.3, 0.3])
    ax3 = fig.add_axes([0.1, 0.15, 0.3, 0.3])
    ax4 = fig.add_axes([0.45, 0.15, 0.3, 0.3])
    ax5 = fig.add_axes([0.82, 0.65, 0.015, 0.13])
    ax6 = fig.add_axes([0.82, 0.45, 0.015, 0.13])
    ax7 = fig.add_axes([0.82, 0.25, 0.015, 0.13])
    axs = [ax1, ax2, ax3, ax4, ax5, ax6, ax7]
    return fig, axs


def heatmap(data, row_labels, col_labels, x_label, y_label,
            ax, title, labelright=True, **kwargs):
    # Plot the heatmap
    im = ax.imshow(data, **kwargs)
    ax.yaxis.tick_right()

    # Show all ticks and label them with the respective list entries.
    ax.set_xticks(np.arange(data.shape[1]), labels=col_labels, rotation=45)
    ax.set_yticks(np.arange(data.shape[0]), labels=row_labels)

    ax.set_xlabel(x_label, fontdict=LABEL_FONTDICT)
    ax.set_ylabel(y_label, fontdict=LABEL_FONTDICT)

    # Let the horizontal axes labeling appear on top.
    ax.tick_params(top=False, bottom=False, left=False, right=False,
                   labelbottom=True, labelright=labelright)

    # Create white grid.
    ax.set_xticks(np.arange(data.shape[1] + 1) - .5, minor=True)
    ax.set_yticks(np.arange(data.shape[0] + 1) - .5, minor=True)
    ax.grid(which="minor", color="w", linestyle='-', linewidth=3)
    ax.tick_params(which="minor", bottom=False, right=False)

    # Set title.
    ax.set_title(title, fontdict=NUMBER_TITLE_FONTDICT, position=[-0.15, 0])

    return im


def add_colorbar(fig, im, cax, title):
    cb = fig.colorbar(im, cax=cax)
    cb.ax.set_title(title, loc="left", fontsize=10, weight='bold')
    cb.ax.tick_params(labelsize=10, direction='in')
    return cb


def line_chart(x, y, x_label, y_label, color, ax, title):
    im = ax.plot(x, y, 'o-', color=color)
    ax.set_xlabel(x_label, fontdict=LABEL_FONTDICT)
    ax.set_ylabel(y_label, fontdict=LABEL_FONTDICT)
    ax.set_title(title, fontdict=NUMBER_TITLE_FONTDICT, position=[-0.15, 0])
    return im


def preproc_max_inhibition_rate_df(max_inhibition_rate_df):
    max_value = max_inhibition_rate_df.max().max()
    if float(max_value) > 0.1:
        min_value = max_value - 0.1
        return max_inhibition_rate_df[max_inhibition_rate_df > min_value]
    else:
        return max_inhibition_rate_df


def plot_pipeline(bioactivity_df, block_score_df, max_inhibition_rate_df, relative_signal_intensity_df, output_folder):
    plt.rcParams['font.sans-serif'] = ['Times New Roman']
    plt.rcParams['font.weight'] = "bold"
    fig, axs = set_layout()

    im1 = line_chart(x=bioactivity_df.index, y=bioactivity_df.values,
                     x_label=FIG_A_XLABEL, y_label=FIG_A_YLABEL,
                     color=FIG_A_COLOR, ax=axs[0], title="A")

    im2 = heatmap(block_score_df.values.T,
                  row_labels=block_score_df.columns.get_level_values(1), col_labels=block_score_df.index,
                  x_label=FIG_BCD_XLABEL, y_label=FIG_BCD_YLABEL,
                  ax=axs[1], title="B", cmap=FIG_B_COLOR)

    max_inhibition_rate_df = preproc_max_inhibition_rate_df(max_inhibition_rate_df)
    im3 = heatmap(max_inhibition_rate_df.values.T,
                  row_labels=max_inhibition_rate_df.columns.get_level_values(1),
                  col_labels=max_inhibition_rate_df.index,
                  x_label=FIG_BCD_XLABEL, y_label=FIG_BCD_YLABEL, labelright=False,
                  ax=axs[2], title="C", cmap=FIG_C_COLOR)

    im4 = heatmap(relative_signal_intensity_df.values.T,
                  row_labels=relative_signal_intensity_df.columns.get_level_values(1),
                  col_labels=relative_signal_intensity_df.index,
                  x_label=FIG_BCD_XLABEL, y_label=FIG_BCD_YLABEL,
                  ax=axs[3], title="D", cmap=FIG_D_COLOR)

    add_colorbar(fig, im2, cax=axs[4], title=FIG_B_CBAR_TITILE)
    add_colorbar(fig, im3, cax=axs[5], title=FIG_C_CBAR_TITILE)
    add_colorbar(fig, im4, cax=axs[6], title=FIG_D_CBAR_TITILE)

    plt.savefig(os.path.join(output_folder, M3_FIG_NAME), dpi=300)

    # plt.show()


# if __name__ == '__main__':
#     fp = "/Users/zhouzhenyi/Documents/github/SciProc/MSBFF/msbff/test/rawdata.csv"
#     _, _, _, bioactivity_df = read_rawdata(fp)
#
#     csv_path = "/Users/zhouzhenyi/Documents/github/SciProc/MSBFF/msbff/test/dataextraction.csv"
#     df = pd.read_csv(csv_path)
#
#     block_score_df, max_inhibition_rate_df, relative_signal_intensity_df = processing_pipeline(df, 0, 8, 1, 0, 800, 100)
#     # print(block_score_df.columns.get_level_values(1))
#     plot_pipeline(bioactivity_df, block_score_df, preproc_max_inhibition_rate_df(max_inhibition_rate_df),
#                   relative_signal_intensity_df, ".")



#!/usr/bin/env python
from os import path

from setuptools import setup


def get_version(rel_path):
    """get version information from __init__.py file."""
    with open(rel_path) as fp:
        cont = fp.read().splitlines()
    for line in cont:
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version infomation.")


curr_dir = path.dirname(path.abspath(__file__))
version = get_version(path.join(curr_dir, 'msbff', '__init__.py'))

setup(
    name='msbff',
    version=version,
    url='https://github.com/BioGavin/msbff',
    license='GPL',
    author='Zhen-Yi Zhou',
    author_email="gavinchou64@gmail.com",
    description="Bioactive Fractions Filtering.",
    classifiers=["Programming Language :: Python :: 3.8",
                 "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
                 "Operating System :: OS Independent"],
    scripts=['scripts/msbff'],
    packages=['msbff'],
    python_requires=">=3.8",
    install_requires=[
        "pandas",
        "matplotlib",
        "numpy"
    ]
)
